//     ========================== ListInterface.h ===========================
//     **********************************************************************
//     Author: Yeremeyev Dmitriy Vladimirovich *** mailto:edvmail@mail.ru
//     Last update : 10.11.2009
//     **********************************************************************

#if !defined(_ListInterface_H_INCLUDED_)
#define _ListInterface_H_INCLUDED_

// Macros definition: MIN/MAX, NULL
#ifndef MIN
#define MIN(a,b)  ((a) > (b) ? (b) : (a))
#endif

#ifndef MAX
#define MAX(a,b)  ((a) < (b) ? (b) : (a))
#endif

#ifndef NULL
#define NULL 0
#endif

//========================================================================================
//================================== Class: iList ========================================
//========================================================================================

// List data structure
template <class iData>
struct iListItem {
	iData d;          // Data
	iListItem* next;  // Pointer to next item (it equal NULL if end of list)
	iListItem* prev;  // Pointer to previous item
};

// Template of data list
template <class iData>
class iList {
public:

	// Definition of data list
	typedef iListItem<iData> iListItem;

private:

	int  mTotal; // Total number of list items
	iListItem *mpFirst, *mpActual, *mpLast; // Pointers to the list items

	// Reset this copy to the start point without memory release
	inline void Reset() {
		mpFirst = NULL; mpActual = NULL; mpLast = NULL;
		mTotal = 0;
	}

public:

	// Class constructor
	iList() { Reset(); }

	// Class destructor
	~iList() { Clear(); }

	// Constructor of copying
	iList(iList<iData>& aList) {
		// Copying an image of data
		mTotal   = aList.mTotal;
		mpFirst  = aList.mpFirst;
		mpActual = aList.mpActual;
		mpLast   = aList.mpLast;
		aList.Reset();
	}

	// Redefinition of operator +=
	const iList& operator+=(const iList& aList) {
		iListItem *p = aList.mpFirst;
		while(p) { Append(p->d); p = p->next; }
		return *this;
	}

	// Redefinition of assignment operator
	const iList& operator=(const iList& aList) {
		Clear();
		*this += aList;
		return *this;
	}

	// Get total number of items
	int Total() const { return mTotal; }
	
	// Append new item
	iListItem* Append(const iData& aData) {
		
		iListItem* p = new iListItem;
		p->d = aData;
		p->next = NULL;
		if(!mpFirst) { mpFirst = p; }

		if(mpLast)   {
			p->prev = mpLast;
			mpLast->next = p;
		} else { p->prev = NULL; }
		
		mpLast = p; mTotal++;

		return p;
	}

	// Insert new item before actual item
	iListItem* InsertBefore(const iData& aData) {
		if(!mpFirst) { return Append(aData); }

		iListItem* p = new iListItem;
		p->d = aData;

		if(mpActual) {
			if(mpActual->prev) { mpActual->prev->next = p; }
			p->prev = mpActual->prev; mpActual->prev = p;
			p->next = mpActual;
			if(mpActual == mpFirst) { mpFirst = p; }
		} else { p->prev = NULL; p->next = mpFirst; mpFirst = p; }

		mTotal++;

		return p;
	}

	// Insert new item with increase sorting
	iListItem* InsertBySorting(const iData& aData, iListItem* apStartFromPos = NULL) {
		if(!mpFirst) { return Append(aData); }

		iData* pData;

		if(apStartFromPos) {
			SetActualPosition(apStartFromPos);
		} else {
			GoAtStart();
		}

		while(pData = GetNext()) {
			if(*pData > aData) { return InsertBefore(aData); }
		}
		return Append(aData);
	}

	// Set the first item as actual
	void GoAtStart() { mpActual = NULL; }

	// Set the last item as actual
	void GoAtEnd() {  mpActual = mpLast; }

	// Get an actual item position in list
	iListItem* GetActualPosition() const { return mpActual; }

	// Set the actual item position in list
	void SetActualPosition(iListItem* apPosition) {	mpActual = apPosition; }

	// Write a data to actual item position
	void Write(const iData& aData) {
		if(mpActual) { mpActual->d = aData;	}
	}

	// Get pointer to data from actual item position
	iData* Read() {
		if(mpActual) { return &mpActual->d;	} else { return NULL; }
	}

	// Get pointer to data at next list item
	iData* GetNext() {
		if(mpActual) {
			if(!mpActual->next) { return NULL; }

			mpActual = mpActual->next;
		} else { mpActual = mpFirst; }
		
		return &mpActual->d;
	}

	// Get pointer to data at previous list item
	iData* GetPrev() {
		if(!mpActual) { return NULL; }

		iListItem *pActual = mpActual;
		mpActual = mpActual->prev;
		
		return &pActual->d;
	}

	// Get pointer to first item in list
	iListItem* GetFirstPosition() const { return mpFirst; }

	// Get pointer to data at next list item
	static iData* GetNext(iListItem** appLI) {
		if(!appLI || !(*appLI)) { return NULL; }
		iData* p = &(*appLI)->d;
		*appLI = (*appLI)->next;
		return p;
	}
		
	// Clearing of list data
	void Clear() {
		iListItem *remove, *p = mpFirst;

		while(p) {	remove = p; p = p->next; delete remove; }
		Reset();
	}

	// Delete list item
	void Delete(iListItem* apDelItem) {
		if(apDelItem) {
			if(apDelItem == mpFirst) {
				
				if(apDelItem == mpLast) { mpLast = NULL; }

				mpFirst  = apDelItem->next;
				if(apDelItem->next) { apDelItem->next->prev = NULL; }
				
			} else {
				if(apDelItem == mpLast) {
					mpLast = apDelItem->prev;
					mpLast->next = NULL;
				} else {
					apDelItem->prev->next = apDelItem->next;
					apDelItem->next->prev = apDelItem->prev;
				}
			}
			delete apDelItem; mTotal--;
		}
	}

	// Delete actual list item
	void DeleteActual() {
		if(mpActual) {
			iListItem* pDelItem = mpActual;
			mpActual = mpActual->next;
			Delete(pDelItem);
		}
	}
};

// Template of list comparison at equality
template <class iData, class iListItem>
bool ListsIsEqual(const iList<iData>& aL1, const iList<iData>& aL2, iListItem*) {
	if(aL1.Total() != aL2.Total()) { return false; }

	iListItem* pLi1 = aL1.GetFirstPosition();
	iListItem* pLi2 = aL2.GetFirstPosition();

	iData *pD1, *pD2;

	// Compare item by item
	while(pD1 = iList<iData>::GetNext(&pLi1)) {
		pD2 = iList<iData>::GetNext(&pLi2);
		if(*pD1 != *pD2) { return false; }
	}

	return true;
}

// Redefinition of operator "equal" for class iList
template <class iData>
bool operator==(const iList<iData>& aL1, const iList<iData>& aL2)
{
	return ListsIsEqual(aL1, aL2, (iList<iData>::iListItem*)NULL);
}

// Redefinition of operator "bigger" for class iList
template <class iData>
bool operator>(const iList<iData>& aL1, const iList<iData>& aL2)
{
	return (aL1.Total()>>1) > (aL2.Total()>>1);
}

// Redefinition of operator "smaller" for class iList
template <class iData>
bool operator<(const iList<iData>& aL1, const iList<iData>& aL2)
{
	return (aL1.Total()>>1) < (aL2.Total()>>1);
}

//========================================================================================
//============================== Class: iAssociatedList ==================================
//========================================================================================

// Data structure of associated list
template <class iKey, class iData>
struct iAscData {
	iKey  Key;
	iData Data;
	void* pIdx;
};

// Template of associated list
template <class iKey, class iData>
class iAssociatedList {
public:

	// Data definition for associated list
	typedef iAscData<iKey, iData> iAscData;

	// Definition of pair list
	typedef iList<iAscData> iPairList;

	// Item definition for associated list
	typedef iListItem<iAscData> iAscItem;

	// Function definition for event "if key is found"
	typedef void (*TYPE_IfKeyFound)(iPairList* apPL, const iAscData& aData, void* args);
	typedef TYPE_IfKeyFound iIfKeyFound;

private:

	#define cLI_IndexOrder 10 // Indexation for each ten pairs
	
	// Structure of index data for associated list
	struct iIndexItem {
		iKey  Key;
		iAscItem *Pos;
	};
	
	// Definition of indexation list
	typedef iList<iIndexItem> iIndexList;
	typedef iListItem<iIndexItem> iIndexListItem;

	double mKeyMin, mKeyMax; // Minimal/maximal key value
	iPairList*  mpPairList;  // Pointer to pair list
	iIndexList* mpIndexList; // Pointer to indexation list

	// Definition of function "compare keys"
	typedef bool (*TYPE_Compare)(const iKey& k1, const iKey& k2, void* args);
	typedef TYPE_Compare iCompare;

	// Definition of function "compare keys by similarity"
	static bool doSimCompare(const iKey& k1, const iKey& k2, void* args) {
		double* pSimilarityThreshold = (double*)args;
		return iKey::Similarity(k1, k2) >= *pSimilarityThreshold;
	}

	// Definition of function "compare keys by identically"
	static bool doIdenticalCompare(const iKey& k1, const iKey& k2, void*) {
		return k1 == k2;
	}

	// Find data in associated list
	iData* Find(const iKey& aKey, iCompare FuncCompare,
				double* apSimTrd = NULL, iIndexListItem* apStartFrom = NULL) {

		iAscData  *pLI;
		iIndexItem *pIdxDat;
		iData *pData = NULL;
		iPairList::iListItem *pFrom = NULL, *pTo = NULL;

		if(!mpPairList->Total()) { return NULL; }

		double KeyBase = mKeyMax - mKeyMin;

		if(aKey > mKeyMin && KeyBase) {
		
			double InKey = (aKey - mKeyMin)/KeyBase;
			double InKeyFrom = (apSimTrd)?(InKey + *apSimTrd - 1.0):InKey;
			double InKeyTo   = (apSimTrd)?(InKey - *apSimTrd + 1.0):InKey;

			// Define search limitation
			if(apStartFrom) {
				mpIndexList->SetActualPosition(apStartFrom);
				pFrom = apStartFrom->d.Pos;
			} else {
				mpIndexList->GoAtStart();
			}

			while(pIdxDat = mpIndexList->GetNext()) {
				double ActKey = (pIdxDat->Key - mKeyMin)/KeyBase;
				if(InKeyTo < ActKey) {
					pTo = pIdxDat->Pos; break;
				} else {
					if(InKeyFrom > ActKey) { pFrom = pIdxDat->Pos; }
				}
			}

			// Set the start position for searching
			mpPairList->SetActualPosition(pFrom);
		} else { mpPairList->GoAtStart(); }

		while((pLI = mpPairList->GetNext()) && mpPairList->GetActualPosition() != pTo) {
			// Is similarity threshold exceeded?
			if(FuncCompare(pLI->Key, aKey, apSimTrd)) {// Yes
				// Data is found
				pData = &pLI->Data; break;
			}
		}

		return pData;
	}
	
public:

	// Class constructor
	iAssociatedList() {
		mpPairList  = new iPairList;
		mpIndexList = new iIndexList;
	}

	// Class destructor
	~iAssociatedList() {
		if(mpPairList)  { delete mpPairList;  mpPairList  = NULL; }
		if(mpIndexList) { delete mpIndexList; mpIndexList = NULL; }
	}

	// Constructor of copying
	iAssociatedList(iAssociatedList<iKey, iData>& aList) {
		// Copying an image of data
		mpPairList  = aList.mpPairList;  aList.mpPairList  = NULL;
		mpIndexList = aList.mpIndexList; aList.mpIndexList = NULL;
	}

	// Redefinition of assignment operator
	const iAssociatedList& operator=(const iAssociatedList& aList) {
		iAscData* pData;
		iAscItem* pLI = aList.mpPairList->GetFirstPosition();
		mpIndexList->Clear();
		mpPairList->Clear();

		for(int i=0; pData = iAssociatedList::GetNext(&pLI); i++) {
			double dKey = double(pData->Key);
			if(!i) {
				mKeyMin = dKey;
				mKeyMax = dKey;
			} else {
				if(mKeyMin > dKey) { mKeyMin = dKey; }
				if(mKeyMax < dKey) { mKeyMax = dKey; }
			}
			
			iAscItem* pNew = mpPairList->Append(*pData);

			if(!(i % cLI_IndexOrder)) {
				iIndexItem IdxDat;
				IdxDat.Key = pNew->d.Key;
				IdxDat.Pos = pNew;
				pNew->d.pIdx = mpIndexList->Append(IdxDat);
			} else {
				pNew->d.pIdx = NULL;
			}
		}

		return *this;
	}

	// Addition of pair (key - data)
	void AddPair(const iKey& aKey, const iData& aData,
				 iIfKeyFound FuncIfKeyFound = NULL, void* args = NULL) {

		iAscData d;
		d.Key  = aKey;
		d.Data = aData;
		d.pIdx = NULL;

		double dKey = double(aKey);
		if(!mpPairList->Total()) {
			mKeyMin = dKey;
			mKeyMax = dKey;
		} else {
			if(mKeyMin > dKey) { mKeyMin = dKey; }
			if(mKeyMax < dKey) { mKeyMax = dKey; }
		}

		iIndexItem *pIdxDat;
		iIndexListItem* pStartFrom = NULL;

		// Define search limitation
		mpIndexList->GoAtStart();
		while(pIdxDat = mpIndexList->GetNext()) {
			if(pIdxDat->Key > aKey) { break; }
			pStartFrom = mpIndexList->GetActualPosition();
		}

		// Is key found?
		if(Find(aKey, doIdenticalCompare, NULL, pStartFrom)) {// Yes
			if(FuncIfKeyFound) {
				FuncIfKeyFound(mpPairList, d, args);
			} else {
				mpPairList->Write(d);
			}
			return;
		}

		// Adding of new item to the pair list
		iPairList::iListItem* pPL = mpPairList->InsertBySorting(d, pStartFrom?pStartFrom->d.Pos:NULL);

		// Add a new index pointer?
		if(!(mpPairList->Total() % cLI_IndexOrder) || mpPairList->Total() == 1) {// Yes
			iIndexItem IdxDat;
			IdxDat.Key = pPL->d.Key;
			IdxDat.Pos = pPL;
			pPL->d.pIdx = mpIndexList->Append(IdxDat);
		} else {
			pPL->d.pIdx = NULL;
		}
	}


	// Deleting of pair out from list
	void DeletePair(const iKey& aKey) {
		if(Find(aKey, doIdenticalCompare)) {
			iAscItem* pFoundItem = mpPairList->GetActualPosition();

			if(pFoundItem->d.pIdx) {
				mpIndexList->Delete((iIndexListItem*)pFoundItem->d.pIdx);
			}

			mpPairList->DeleteActual();
		}
	}

	// Find data in associated list by similarity threshold of keys
	iData* Find(const iKey& aKey, double aSimilarityThreshold) {
		return Find(aKey, doSimCompare, &aSimilarityThreshold);
	}

	// Find data in associated list by identical of keys
	iData* Find(const iKey& aKey) {
		return Find(aKey, doIdenticalCompare);
	}

	// Get pointer to first item in list
	iAscItem* GetFirstPosition() const { return mpPairList->GetFirstPosition(); }

	// Get pointer to data at next list item
	static iAscData* GetNext(iAscItem** appLI) {
		return iPairList::GetNext(appLI);
	}

	// Get total number of items
	int Total() const { return mpPairList->Total(); }

	// Clearing of list data
	void Clear() {
		if(mpPairList)  { mpPairList->Clear(); }
		if(mpIndexList) { mpIndexList->Clear(); }
	}

	// Get pointer to the pair list
	iPairList* GetPairListPtr() const { return mpPairList; }
};

// Redefinition of operator "bigger" for structure iAscData<iKey, iData>
template <class iKey, class iData>
bool operator>(const iAscData<iKey, iData>& d1,
			   const iAscData<iKey, iData>& d2) {
	return d1.Key > d2.Key;
}

// Redefinition of operator "not equal" for structure iAscData<iKey, iData>
template <class iKey, class iData>
bool operator!=(const iAscData<iKey, iData>& d1,
				const iAscData<iKey, iData>& d2)
{
	return d1.Key != d2.Key;
}

// Redefinition of operator "equal" for class iAssociatedList
template <class iKey, class iData>
bool operator==(const iAssociatedList<iKey, iData>& aL1,
				const iAssociatedList<iKey, iData>& aL2)
{
	return ListsIsEqual(*aL1.GetPairListPtr(), *aL2.GetPairListPtr(),
						(iAssociatedList<iKey, iData>::iPairList::iListItem*)NULL);
}

// Redefinition of operator "bigger" for class iAssociatedList
template <class iKey, class iData>
bool operator>(const iAssociatedList<iKey, iData>& aL1,
			   const iAssociatedList<iKey, iData>& aL2)
{
	return aL1.GetPairListPtr()->Total() > aL2.GetPairListPtr()->Total();
}

// Redefinition of operator "smaller" for class iAssociatedList
template <class iKey, class iData>
bool operator<(const iAssociatedList<iKey, iData>& aL1,
			   const iAssociatedList<iKey, iData>& aL2)
{
	return aL1.GetPairListPtr()->Total() < aL2.GetPairListPtr()->Total();
}


//========================================================================================
//================================== Class: iIdString ====================================
//========================================================================================

// Identifiers string template
template <class iIdentifierType, class iIdentifierItem>
class iIdString {
	// Definition of identifier list
	typedef iList<iIdentifierItem> iIdentifierList;

	// Definition of identifiers string
	typedef iAssociatedList<iIdentifierType, long> iString;

	long     mIdTotal;  // Identifiers total number
	iString* mpString;  // Pointer to identifiers string list

public:

	// Data definition of identifiers string
	typedef iAscData<iIdentifierType, long> iIdData;

	// Definition of pair list
	typedef iList<iIdData> iPairList;

	// Item definition of identifiers string
	typedef iListItem<iIdData> iIdItem;
	
	// Class constructor
	iIdString() : mIdTotal(0) {
		mpString = new iString;
	}

	// Class destructor
	~iIdString() {
		if(mpString) { delete mpString; mpString = NULL; }
	}

	// Constructor of copying
	iIdString(iIdString<iIdentifierType, iIdentifierItem>& aIdStr) {
		// Copying an image of data
		mIdTotal = aIdStr.mIdTotal; aIdStr.mIdTotal = 0;
		mpString = aIdStr.mpString; aIdStr.mpString = NULL;
	}
	
	// Redefinition of assignment operator
	const iIdString& operator=(const iIdString& aIdStr) {
		// Make data copy
		mIdTotal  = aIdStr.mIdTotal;
		*mpString = *aIdStr.mpString;
		return *this;
	}

	// Redefinition of operator +=
	const iIdString& operator+=(const iIdentifierList& aIdList) {
		Create(aIdList, true);
		return *this;
	}

	// Redefinition of type conversion into "double"
	operator double() const { return Total(); }

	// Creating of identifiers string
	void Create(const iIdentifierList& aIdList, bool aFlagAdd = false) {
		long count = 1;
		iIdentifierItem *actual, *prev = NULL;

		iString::iIfKeyFound pIfKeyFound = NULL;

		if(aFlagAdd) {
			// List data addition to existed string
			pIfKeyFound = doIfKeyFound;
			mIdTotal   += aIdList.Total();
		} else {
			// New string
			if(mpString->Total())  { mpString->Clear(); }
			mIdTotal = aIdList.Total();
		}

		iIdentifierList::iListItem* pIL = aIdList.GetFirstPosition();

		while(actual = iIdentifierList::GetNext(&pIL)) {
			if(prev) {
				// Is identifier repeating?
				if(actual->Identifier != prev->Identifier) {// No
					mpString->AddPair(prev->Identifier, count, pIfKeyFound);
					count = 1;
				} else { count++; }
			}
			prev = actual;
		}
		if(prev) { mpString->AddPair(prev->Identifier, count, pIfKeyFound); }
	}

	// Get similarity between two strings
	static double Similarity(const iIdString<iIdentifierType, iIdentifierItem>& aIdStr1,
							 const iIdString<iIdentifierType, iIdentifierItem>& aIdStr2) {
		
		double IdTotalMax  = MAX(aIdStr1.mIdTotal, aIdStr2.mIdTotal);

		if(!IdTotalMax) { return 1.0; }

		// Is in first string contained more characters than in second string?
		bool IdStr1IsBiger = aIdStr1.Total() > aIdStr2.Total();
		
		double IdCoincidence = 0;
		iIdItem* pPair = ((IdStr1IsBiger)?aIdStr2:aIdStr1).GetFirstPosition();
		iIdData* pActualId;
		// Searching of match
		while(pActualId = GetNext(&pPair)) {
			long* pCount = ((IdStr1IsBiger)?aIdStr1:aIdStr2).mpString->Find(pActualId->Key);
			// Is identifiers matched?
			if(pCount) { // Yes
				IdCoincidence += MIN(*pCount, pActualId->Data);
			}
		}

		return IdCoincidence/IdTotalMax;
	}

	// Function for event "if key is found"
	static void doIfKeyFound(iPairList* apPL, const iIdData& aData, void* args) {
		// Get data from actual position
		iIdData* pActualData = apPL->Read();
		// Increase a character counter
		if(pActualData) {
			pActualData->Data += aData.Data;
		}
	}

	// Get pointer to first item in list
	iIdItem* GetFirstPosition() const { return mpString->GetFirstPosition(); }

	// Get pointer to data at next list item
	static iIdData* GetNext(iIdItem** appLI) {
		return iString::iPairList::GetNext(appLI);
	}

	// Get total number of unique identifiers in list
	int Total() const { return mpString->Total(); }

	// Get pointer to identifiers string list
	iString* GetStringPtr() const { return mpString; }
};

// Redefinition of operator "equal" for class iIdString
template <class iIdentifierType, class iIdentifierItem>
bool operator==(const iIdString<iIdentifierType, iIdentifierItem>& s1,
				const iIdString<iIdentifierType, iIdentifierItem>& s2)
{
	return *s1.GetStringPtr() == *s2.GetStringPtr();
}

// Redefinition of operator "bigger" for class iIdString
template <class iIdentifierType, class iIdentifierItem>
bool operator>(const iIdString<iIdentifierType, iIdentifierItem>& s1,
				const iIdString<iIdentifierType, iIdentifierItem>& s2)
{
	return *s1.GetStringPtr() > *s2.GetStringPtr();
}

// Redefinition of operator "smaller" for class iIdString
template <class iIdentifierType, class iIdentifierItem>
bool operator<(const iIdString<iIdentifierType, iIdentifierItem>& s1,
				const iIdString<iIdentifierType, iIdentifierItem>& s2)
{
	return *s1.GetStringPtr() < *s2.GetStringPtr();
}


#endif // !defined(_ListInterface_H_INCLUDED_)