//     ========================== Navigator.cpp =============================
//     **********************************************************************
//     Author: Yeremeyev Dmitriy Vladimirovich *** mailto:edvmail@mail.ru
//     Last update : 10.01.2010
//     **********************************************************************

#include "Navigator.h"
#include <windows.h>

#define cCmdPauseTime 1000 // Time of pause between commands

NAVIGATOR_API TCmdArray* gpCmdArray;         // Robot command buffer
NAVIGATOR_API bool gNovaGateMode  = false;   // "Nova gate" mode flag
NAVIGATOR_API ushort gCheckpointCounter = 0; // Counter of checkpoint setting
NAVIGATOR_API int gWayIdx = 0;               // Route index

TNavigateData gNvg;             // Global navigation data
HANDLE hCmdExchangeSection = 0; // Handle of exchange section of commands


// Starting of navigation mode
NAVIGATOR_API void StartNavigateMode(bool aNovaGateMode)
{
	gNovaGateMode = aNovaGateMode;
	
	if(gpCmdArray) {
		memset(gpCmdArray, 0, sizeof(TCmdArray));
		return;
	}
	
	// Initialization of exchange section
	hCmdExchangeSection = CreateFileMapping(INVALID_HANDLE_VALUE,
		NULL, PAGE_READWRITE, 0, sizeof(TCmdArray), "RobotCmdExchange");
	TCmdArray* pCmdArray = (TCmdArray*) MapViewOfFile(hCmdExchangeSection,FILE_MAP_ALL_ACCESS,0,0,sizeof(TCmdArray));
	// Sequence of start commands for robot
	memset(&gNvg, 0, sizeof(TNavigateData));
	pCmdArray->Cmd[cmFIRE]  = true;
	Sleep(cCmdPauseTime);
	pCmdArray->Cmd[cmPOWER] = true;
	Sleep(cCmdPauseTime);
	pCmdArray->Cmd[cmFIRE]  = false;
	pCmdArray->Cmd[cmPOWER] = false;
	Sleep(cCmdPauseTime);
	pCmdArray->Cmd[cmFIRE] = true;
	Sleep(cCmdPauseTime/2);
	pCmdArray->Cmd[cmFIRE] = false;
	gpCmdArray = pCmdArray;
}

// Ending of navigation mode
NAVIGATOR_API void FinishNavigateMode()
{
	gNovaGateMode = false;
	// Closing of exchange section
	if(gpCmdArray) {
		TCmdArray* pCmdArray = gpCmdArray; gpCmdArray = NULL;
		memset(pCmdArray, 0, sizeof(TCmdArray));
		pCmdArray->Cmd[cmPOWER] = true;
		Sleep(cCmdPauseTime);
		pCmdArray->Cmd[cmPOWER] = false;			
		UnmapViewOfFile(pCmdArray);
	}
	if(hCmdExchangeSection) { CloseHandle(hCmdExchangeSection); }
}

// Turn the robot body on the spot
void TurnBody(ushort aCounter, bool aTurnLeft = true)
{
	bool* pTurnCmd;
	
	if(aTurnLeft) {
		pTurnCmd = &gpCmdArray->Cmd[cmLEFT];
		gpCmdArray->Cmd[cmRIGHT] = false;
	} else {
		pTurnCmd = &gpCmdArray->Cmd[cmRIGHT];
		gpCmdArray->Cmd[cmLEFT] = false;
	}
	// Turn the body
	if(aCounter & 0x08) {
		if(aCounter & 0x04) {
			gpCmdArray->Cmd[cmFORWARD] = true;
			gpCmdArray->Cmd[cmBACKWARDS] = false;
		} else {
			gpCmdArray->Cmd[cmFORWARD] = false;
			gpCmdArray->Cmd[cmBACKWARDS] = true;
		}
		*pTurnCmd = true;
	} else {
		// Pause
		*pTurnCmd = false;
		gpCmdArray->Cmd[cmFORWARD] = false;
		gpCmdArray->Cmd[cmBACKWARDS] = false;
	}
}

// Random motion on the spot
void RandomMotion()
{
	if(!(rand()%8)) {
		// Turn the robot body to the left side by one impulse
		gpCmdArray->Cmd[cmLEFT] = true;
		gpCmdArray->Cmd[cmFORWARD] = true;
	} else {
		if(!(rand()%8)) {
			// Turn the robot body to the right side by one impulse
			gpCmdArray->Cmd[cmRIGHT] = true;
			gpCmdArray->Cmd[cmBACKWARDS] = true;
		} else {
			gpCmdArray->Cmd[cmLEFT] = false;
			gpCmdArray->Cmd[cmRIGHT] = false;
			gpCmdArray->Cmd[cmFORWARD] = false;
			gpCmdArray->Cmd[cmBACKWARDS] = false;
		}
	}
}

// Robot navigation function
NAVIGATOR_API void Navigate(IplImage* apGray, IplImage* apScreen, TRcgObj* apRcgObj,
							TSetCheckpointFunc SetCheckpoint, TLogMessageFunc LogMessage) {

	static ushort PhaseCounter = 0;
	static bool   TurnLeft = true;

	if(gpCmdArray) {
		if(!gNovaGateMode) {
			// Get previous state of robot control signals
			bool MvFrw = gpCmdArray->Cmd[cmFORWARD];
			bool MvBkw = gpCmdArray->Cmd[cmBACKWARDS];
			bool MvLft = gpCmdArray->Cmd[cmLEFT];
			bool MvRht = gpCmdArray->Cmd[cmRIGHT];
			bool TrLft = gpCmdArray->Cmd[cmTURRET_LEFT];
			bool TrRht = gpCmdArray->Cmd[cmTURRET_RIGHT];
			// Clear the command exchange buffer
			memset(gpCmdArray, 0, sizeof(TCmdArray));
			// Is some object found?
			if(apRcgObj->ObjTotal) {// Yes
				CvRect ObjRect = apRcgObj->ObjRect[0];
				long ObjIdxMax = -1;
				// Find object with maximal index
				for(int i=0; i<apRcgObj->ObjTotal; i++) {
					if(ObjIdxMax < apRcgObj->ObjIdx[i]) {
						ObjIdxMax = apRcgObj->ObjIdx[i];
						ObjRect = apRcgObj->ObjRect[i];
					}
				}
				// Get base size of object/screen
				double ScrBase = MIN(apGray->width, apGray->height);
				double ObjBase = MAX(ObjRect.width, ObjRect.height);
				// Get size ratio of object/screen
				double ObjZoom = ObjBase/ScrBase;
				// Calculating of difference between center of object and screen
				int cXs = apGray->width>>1;
				int cXo = ObjRect.x + (ObjRect.width>>1);
				int cXd = cXs - cXo;
				// Does robot see large object (it too close)?
				if(ObjZoom > 0.6  || (ObjZoom > 0.57 && MvBkw)) {// Yes
					// Robot will rolled away back
					gpCmdArray->Cmd[cmBACKWARDS] = true;
				}
				
				// Does robot see small object (it too far)?
				if(ObjZoom < 0.37 || (ObjZoom < 0.42 && MvFrw)) {// Yes
					// Robot will come nearer
					gpCmdArray->Cmd[cmFORWARD]   = true;
				}
				
				double CentralRate = fabs(double(cXd))/double(cXs);
				
				// Is the object at left side from robot?
				if(cXd > 40) {// Yes
					// Turn the robot body to the left side by one impulse
					if((!MvLft || CentralRate > 0.33) && gNvg.TurretBalance > 0) {
						gpCmdArray->Cmd[cmLEFT] = true;
					}
					// Turn the turret to the left
					if(!TrLft) {
						gpCmdArray->Cmd[cmTURRET_LEFT] = true;
						gNvg.TurretBalance++;
					}
				}
				// Is the object at right side from robot?
				if(cXd < -40) {// Yes
					// Turn the robot body to the right side by one impulse
					if((!MvRht || CentralRate > 0.33) && gNvg.TurretBalance < 0) {
						gpCmdArray->Cmd[cmRIGHT] = true;
					}
					// Turn the turret to the right
					if(!TrRht) {
						gpCmdArray->Cmd[cmTURRET_RIGHT] = true;
						gNvg.TurretBalance--;
					}
				}
				gNvg.MotionSwitch = 7;
			} else {// The object is lost from field of vision
				// Is checkpoint found?
				if(!apRcgObj->CheckpointFound) {// Yes
					// Does robot see "gate"?
					if(apRcgObj->GateHorizon) {// Yes
						// Is the gate at left side from robot?
						if(apRcgObj->GateHorizon > 0.09) {// Yes
							TurnLeft = true;
							// Turn the robot body to the left side by one impulse
							if((!MvLft || fabs(apRcgObj->GateHorizon) > 0.195) && gNvg.TurretBalance > 0) {
								gpCmdArray->Cmd[cmLEFT] = true;
							}
							// Turn the turret to the left
							if(!TrLft) {
								gpCmdArray->Cmd[cmTURRET_LEFT] = true;
								gNvg.TurretBalance++;
							}
						}
						// Is the gate at right side from robot?
						if(apRcgObj->GateHorizon < -0.09) {// Yes
							TurnLeft = false;
							// Turn the robot body to the right side by one impulse
							if((!MvRht || fabs(apRcgObj->GateHorizon) > 0.195) && gNvg.TurretBalance < 0) {
								gpCmdArray->Cmd[cmRIGHT] = true;
							}
							// Turn the turret to the right
							if(!TrRht) {
								gpCmdArray->Cmd[cmTURRET_RIGHT] = true;
								gNvg.TurretBalance--;
							}
						}
						// Robot will come nearer
						gpCmdArray->Cmd[cmFORWARD] = true;
					} else {
						// Robot does not see "gate"
						if(PhaseCounter & 0x30) {
							// Random motion on the spot
							RandomMotion();
						} else {
							// Turn the robot body on the spot
							TurnBody(PhaseCounter, TurnLeft);
						}
						
						// Turret alignment
						if(gNvg.TurretBalance) {
							if(gNvg.MotionSwitch) {
								if(gNvg.TurretBalance > 0) {
									gpCmdArray->Cmd[cmLEFT] = true;
								} else {
									gpCmdArray->Cmd[cmRIGHT] = true;
								}
								// Twitch forward-back to align turret
								if(gNvg.MotionSwitch & 2) {
									if(gNvg.MotionSwitch > 3) {
										gpCmdArray->Cmd[cmFORWARD]   = true;
										gpCmdArray->Cmd[cmBACKWARDS] = false;
									} else {
										gpCmdArray->Cmd[cmFORWARD]   = false;
										gpCmdArray->Cmd[cmBACKWARDS] = true;
									}
								} else {
									gpCmdArray->Cmd[cmFORWARD]   = false;
									gpCmdArray->Cmd[cmBACKWARDS] = false;
								}
								gNvg.MotionSwitch--;
							} else {
								gpCmdArray->Cmd[cmLEFT]  = false;
								gpCmdArray->Cmd[cmRIGHT] = false;
							}
							// Turn the turret
							if(gNvg.TurretBalance > 0) {
								gpCmdArray->Cmd[cmTURRET_RIGHT] = true;
								gNvg.TurretBalance--;
							} else {
								gpCmdArray->Cmd[cmTURRET_LEFT] = true;
								gNvg.TurretBalance++;
							}
						}
					}
				}
			}
		} else {// Nova gate mode
			if(gCheckpointCounter) {
				gCheckpointCounter--;

				// Turn the robot body on the spot
				TurnBody(gCheckpointCounter);
				if(!(gCheckpointCounter & 0x0F)) {
					// Setting of checkpoint
					SetCheckpoint(gWayIdx+1, 0xFFFF);
				}
			} else {
				if(!(PhaseCounter & 0x07)) {
					// Setting of checkpoint
					SetCheckpoint(0, 0xFFFF);
				}
			}
		}

		// Control keys showing
		CvScalar bColor = (gNovaGateMode)?cvScalar(0,0,255):cvScalar(0,255,0);
		cvRectangle(apScreen, cvPoint(10, 30), cvPoint(30, 50),
					bColor, gpCmdArray->Cmd[cmLEFT]?CV_FILLED:2,8,0);

		cvRectangle(apScreen, cvPoint(40, 30), cvPoint(60, 50),
					bColor, gpCmdArray->Cmd[cmBACKWARDS]?CV_FILLED:2,8,0);

		cvRectangle(apScreen, cvPoint(70, 30), cvPoint(90, 50),
					bColor, gpCmdArray->Cmd[cmRIGHT]?CV_FILLED:2,8,0);

		cvRectangle(apScreen, cvPoint(40, 60), cvPoint(60, 80),
					bColor, gpCmdArray->Cmd[cmFORWARD]?CV_FILLED:2,8,0);

		cvRectangle(apScreen, cvPoint(10, 110), cvPoint(30, 130),
					bColor, gpCmdArray->Cmd[cmTURRET_LEFT]?CV_FILLED:2,8,0);

		cvRectangle(apScreen, cvPoint(40, 110), cvPoint(60, 130),
					bColor, gpCmdArray->Cmd[cmFIRE]?CV_FILLED:2,8,0);

		cvRectangle(apScreen, cvPoint(70, 110), cvPoint(90, 130),
					bColor, gpCmdArray->Cmd[cmTURRET_RIGHT]?CV_FILLED:2,8,0);
	}
	PhaseCounter--;
}


