//     ======================== RecognitionApp.h ============================
//     **********************************************************************
//     Author: Yeremeyev Dmitriy Vladimirovich *** mailto:edvmail@mail.ru
//     Last update : 03.02.2010
//     **********************************************************************


#if !defined(_RecognitionApp_H_INCLUDED_)
#define _RecognitionApp_H_INCLUDED_

#include "wx/wx.h"
#include <wx/file.h>
#include "AssociativeMemory.h"
#include "ListInterface.h"
#include "AVM_drawing.h"
#include "OpticalFlow.h"
#include "Navigator.h"
#include "cvcam.h"

#define MultilingualSupport(x) x
#define _MS(x) MultilingualSupport(x)
#define _MST(x) MultilingualSupport(_T(x))


//=============================================================================================
//============================= Class: CRecognitionPipeline ===================================
//=============================================================================================

#define cRcgDataFileIdentStrLen 12             // Identifier string length
#define cRcgDataFileIdent       "RcgApp-v0.5"  // Recognition data file identifier string
#define cAMFileName             _T("avm.dat")  // Recognition data file name
#define cObjMaxNum              1000           // Maximal number of recognized objects
#define cObjNameLenMax          64             // Maximal number of characters in object names
#define cBaseKeyImgSize         80             // Base key image size for AVM
#define cPointIdx_s             11             // Checkpoint index in array
#define cHitIdx_c               14             // Index of hit counter in array

// Structure of name table of objects
struct TObjName {
	char s[cObjMaxNum][cObjNameLenMax]; // Object name array
	long ObjIdx;                        // Object index in array
	long Total;                         // Total number of objects in array
};

// Structure of name table of objects
struct TRcgDataFile {
	char     Ident[cRcgDataFileIdentStrLen]; // Object name array
	long     RcgDataSize;                    // Object index in array
	long     ObjDataOffset;                  //    
	TObjName ObjName;                        // Total number of objects in array
};

// Work mode constants
typedef enum {
	cSetTheBaseCoordinateOfObject = 0, // Setting (by user) of a base coordinates of object
	cInputObjectName,                  // Entering of an object name
	cSetTheKeyImageSize,               // Setting of key image size
	cPrepareToTraining,                // Preparing to training
	cTraining,                         // Training mode
	cRecognition                       // Recognition mode
} TWorkMode;

// Video settings constants
typedef enum {
	cDefault = 0,         // Default settings
	cChangeVideoFormat,   // Setting of video format
	cChangeCaptureSource  // Setting of capture source of input video
} TCaptureSettings;


// Recognition pipeline class
class CRecognitionPipeline : public wxStaticText {
public:
	TRcgDataFile             mRcgDataFile;   // Recognition data file
	CvDrawAssociativeTree*   mpDrawTree;     // Pointer to the view class of associative tree
	bool                     mColoredMotion; // Flag of "Colored motion" mode
	CvOpticalFlow*           mpOptFlw;       // Pointer to "Optical flow" class
	CvAssociativeMemory32S*  mpObjAM;        // Pointer to copy of AVM  for objects
	CvAssociativeMemory192U* mpGateAM;       // Pointer to copy of AVM  for gates
	CvSize                   mGateKeySize;   // Image key size of gate

private:	
	CvSize    mImgSize;     // Size of input image	
	TWorkMode mWorkMode;    // Work mode
	IplImage* mpScreen;     // Pointer to the screen memory
	IplImage* mpGray;       // Pointer to the gray-scale input image
	CvRect    mInterestArea;// Interest area (where object located)
	long      mPauseLoop;   // Pause counter before training
	long      mTrainingLoop;// Counter of training loop
	bool      mNeedKeyImageSetting;         // Requirement of key image size setting
	wxString  mHint;                        // Hint string
	bool      mMouseCursorInit;             // Mouse cursor initialization flag
	wxCoord   mMouseCursorX, mMouseCursorY; // Mouse cursor coordinates
	wxSize    mHint_WinSize;                // Size of hint window
	
#define   cHint_OffsetX 20 // Offset of hint window by X axis
#define   cHint_OffsetY 10 // Offset of hint window by Y axis
	CvFont    mHint_Font;      // Font type of hint
	CvSize    mHint_TextSize;  // Text size of hint
	bool      mHint_AtLeft;    // The flag of hint presence at left side of cursor
	double    mAspectRatio;    // Aspect ratio of key image size
	CvSize    mKeyImageSize;   // Key image size
	int       mCamIdx;         // Camera index
	CvFont    mComment_Font;   // Font type of comment
	wxString  mComment;        // Comment string
	int       mCommentTimer;   // Timer of comment showing (on the screen)
	
	wxCriticalSection mProcessing; // Critical section of video processing
	
public:
	// Class constructor
	CRecognitionPipeline(wxWindow* parent, int aCamIdx,
		TCaptureSettings aCaptSet = cDefault, VidFormat* apVidFmt = NULL);
	// Class destructor
	~CRecognitionPipeline();
	// Capture initialization
	wxSize InitCapture(int aCamIdx, TCaptureSettings aCaptSet, VidFormat* apVidFmt);
	// Creation of recognition pipeline
	void CreatePipeline(void* apRcgData = NULL, long aObjDataOffset = 0);
	// Initialization of "gate" memory
	void InitGateMemory();
	// Fitting of gate memory size for new resolution
	void FitGateMemorySize();
	// Reading of recognition data
	void ReadRcgData(void* apRcgData, long aObjDataOffset = 0);
	// Setting of base object coordinates
	void SetTheBaseCoordinateOfObject();
	// Preparing to training
	void PrepareToTraining();
	// Setting of recognition mode
	void SetRecognitionMode();
	// Setting of checkpoint
	void SetCheckpoint(int aWayNum, ushort aGateCode);
	// Handler on "Learn the next object" event
	void OnLearnNextObj(int aWayIdx);
	// Handler on "Clearing of gate data" event
	void OnClearGateData();
	// Handler on "Setting of key image size" event
	void OnSetKeySize();
	// Optimization of associative tree
	void OptimizeAssociativeTree();
	// Writing of recognition data to memory
	void WriteRcgData(void* apRcgData);
	// Get size of recognition data
	long GetRcgDataSize() const;
	// Get offset to the object data
	long GetObjDataOffset() const;
	// Reset the recognition data (as default)
	void RcgDataFileAsDefault();
	// Comment printing at the demo video
	void Comment(wxString aText);
	// Redrawing screen
	void Draw();
	// Handler on "Mouse move" event
	void OnMouseMove(wxMouseEvent& event);
	// Handler on "Left button of mouse down" event
	void OnMouseLeftDown(wxMouseEvent& event);
	// Setting of hint window properties
	void SetHintWin(wxString aTitle, wxSize* apWinSize = NULL, bool aAtLeft = true);
	// Redrawing of hint window
	wxPoint DrawHint(wxCoord x, wxCoord y);
	// Entering to critical section of video processing
	void EnterProcessing() { mProcessing.Enter(); }
	// Leaving from critical section of video processing
	void LeaveProcessing() { mProcessing.Leave(); }
	// Get the index of camera
	int GetCameraIndex() const { return mCamIdx; }
	// Get size of input image
	CvSize GetImgSize() const { return mImgSize; }
	// Callback of recognition pipeline function
	friend void RecognitionPipelineFunc(IplImage* image);

private:
    DECLARE_EVENT_TABLE()
};


//=============================================================================================
//===================================== Class: CMainWnd =======================================
//=============================================================================================

// Main window class
class CMainWnd : public wxFrame
{
	wxString    mDataDir;     // Path to recognition data files
    wxPanel*    mpPanel;      // Panel "Window backing"
    wxListBox*  mpLBoxLog;    // Pointer to the log box
    wxLog*      mpLogTarget;  // Pointer to the log window
	wxBoxSizer* mpRcgPlSizer; // Pointer to window of recognition pipeline
	wxBoxSizer* mpTopSizer;   // Sizer of top window area
	wxBoxSizer* mpBtnSizer;   // Sizer for control buttons
	wxSpinCtrl* mpWayIdx;     // Route index
	wxMenuBar*  mpMenuBar;    // Pointer to main menu
	wxButton*   mpLearnBtn;   // Pointer to button of "learning"
	wxButton*   mpSetWayBtn;  // Pointer to button "Walking by way"

	// Starting of a navigate mode
	void StartNavigateMode(bool aNovaGateMode = false);
	// Finish of a navigate mode
	void FinishNavigateMode();

public:
	// Class constructor
    CMainWnd();
	
    // Loading of recognition data from file
	void* Load(wxString aFileName, long* apObjDataOffset = NULL);
	// Saving of recognition data to file
	void  Save(wxString aFileName);
	// Changing of capture settings
	void ChangeCapture(TCaptureSettings aCaptSet, VidFormat* apVidFmt = NULL);
	// Handler on "Quit" event
    void OnQuit(wxCommandEvent& event);
	// Handler on "Close" event
    void OnClose(wxCloseEvent& event);
    // Handler on "Learning of next object" event
    void OnLearnNextObj(wxCommandEvent& event);
    // Handler on "Walking by way"
    void OnSetWay(wxCommandEvent& event);
	// Handler on "Setting of key image size" event
    void OnSetKeySize(wxCommandEvent& event);
	// Handler on "Showing of input image" event
	void OnShowInputImage(wxCommandEvent& event);
	// Handler on "Writing of demo video" event
	void OnWriteAVIdemo(wxCommandEvent& event);
	// Handler on "Showing of associative tree" event
	void OnShowAssociativeTree(wxCommandEvent& event);
	// Handler on "Nova gate mode" event
	void OnNovaGateMode(wxCommandEvent& event);
	// Handler on "Navigate mode" event
	void OnNavigateMode(wxCommandEvent& event);
	// Handler on "Clearing of gate data" event
	void OnClearGateData(wxCommandEvent& event);
	// Handler on "Showing of motion field" event
	void OnShowMotionField(wxCommandEvent& event);
	// Handler on "Showing of colored motion" event
	void OnShowColoredMotion(wxCommandEvent& event);
	// Handler on "Setting of capture format" event
	void OnCaptureFormat(wxCommandEvent& event);
	// Handler on "Setting of capture source" event
	void OnCaptureSource(wxCommandEvent& event);
	// Handler on "Saving of recognition data to file" event
	void OnSave(wxCommandEvent& event);
	// Handler on "Loading of recognition data from file" event
	void OnLoad(wxCommandEvent& event);
	// Handler on "Showing information about program" event
    void OnMenuAbout(wxCommandEvent& event);
	// Handler on timer event
	void OnTimer(wxTimerEvent& event);

private:
	wxTimer m_Timer;

    DECLARE_EVENT_TABLE()
};

//=============================================================================================
//==================================== Class: LboxLogger ======================================
//=============================================================================================

// Logger box class
class LboxLogger : public wxLog
{
public:
    LboxLogger(wxListBox *lbox, wxLog *logOld)
    {
        m_lbox = lbox;
        m_logOld = logOld;
    }
	
    virtual ~LboxLogger()
    {
        wxLog::SetActiveTarget(m_logOld);
    }
	
private:
    // Information assembly function
    virtual void DoLog(wxLogLevel level, const wxChar *szString, time_t t)
    {
        if ( level == wxLOG_Trace ) {
            if ( m_logOld ) { ((LboxLogger *)m_logOld)->DoLog(level, szString, t); }
        } else {
            wxLog::DoLog(level, szString, t);
        }
    }
	// Making of log string
    virtual void DoLogString(const wxChar *szString, time_t WXUNUSED(t))
    {
		
#ifdef _UNICODE
		int StrLen = wcslen(szString);
#else
		int StrLen = strlen(szString);
#endif		
		wxString msg = _T("");
		for(int i=0; i<StrLen; i++) {
			if(szString[i] == '\n') {
				m_lbox->Append(msg);
				msg = _T("");
			} else { msg += szString[i]; }
		}
		
		m_lbox->Append(msg);
		m_lbox->SetFirstItem(m_lbox->GetCount()-1);
    }
	
    // Pointer to output area
    wxListBox *m_lbox;
	
    // Pointer to previous output area
    wxLog *m_logOld;
};

#endif // !defined(_RecognitionApp_H_INCLUDED_)
